import collections
from typing import Any, List, TypeVar

from rbx.box.fields import merge_pydantic_models

TypeVarT = TypeVar('TypeVarT', bound=Any)


def expand_statements(statements: List[TypeVarT]) -> List[TypeVarT]:
    deg = collections.defaultdict(int)
    dependencies = collections.defaultdict(list)
    for statement in statements:
        if statement.extends is not None:
            deg[statement.name] += 1
            dependencies[statement.extends].append(statement.name)

    # Topological sort.
    #   - We need to expand statements in the order of dependencies.
    #   - This is a simple topological sort.
    #   - If there are multiple statements with indegree 0, we choose the first one.
    st_per_name = {}
    expanded = {}
    st = []
    for statement in statements:
        st_per_name[statement.name] = statement
        if deg[statement.name] == 0:
            st.append(statement)

    while st:
        statement = st.pop()
        expanded_statement = statement.model_copy()
        if statement.extends is not None:
            expanded_statement = merge_pydantic_models(
                expanded[statement.extends], statement
            )

        expanded[statement.name] = expanded_statement

        for dep_name in dependencies[statement.name]:
            deg[dep_name] -= 1
            if deg[dep_name] == 0:
                st.append(st_per_name[dep_name])

    if len(expanded) != len(statements):
        raise ValueError(
            f'Failed to expand statements: only {len(expanded)} out of {len(statements)} were expanded. This means there is a cycle introduced by the `extends` field.'
        )

    return [expanded[statement.name] for statement in statements]



import atexit
import io
import os
import pathlib
import shelve
import shutil
import tempfile
from typing import Any, Dict, List, Optional

from pydantic import BaseModel
from sqlitedict import SqliteDict

from rbx import console
from rbx.grading import grading_context
from rbx.grading.judge.cacher import FileCacher
from rbx.grading.judge.digester import digest_cooperatively
from rbx.grading.judge.storage import copyfileobj
from rbx.grading.profiling import Profiler
from rbx.grading.steps import (
    DigestHolder,
    GradingArtifacts,
    GradingFileOutput,
    GradingLogsHolder,
)

VERBOSE = False


class CacheInput(BaseModel):
    """
    The exact command that was executed, together with
    its set of input and output artifacts.

    This is used as a cache key, which means that, if across
    executions, the command, or the set of artifacts it
    consumes/produces changes, then there will be a cache key
    change, and thus the command will be re-run.
    """

    commands: List[str]
    artifacts: List[GradingArtifacts]
    extra_params: Dict[str, Any] = {}


class CacheFingerprint(BaseModel):
    """
    The state of the artifacts that are not stored in the
    cache key (usually for efficiency/key size reasons), such
    as the hashes of every FS input artifact, or the hashes of
    the produced artifacts.

    This is used for a few things:
        - Check whether the IO files have changed in disk since
          this command was cached, and evict this entry in such case.
        - Check whether the caching storage has changed since this
          command was cached, and evict this entry in such case.
        - Store small side-effects of the cached execution, such as
          execution time, memory, exit codes, etc.
    """

    digests: List[Optional[str]]
    fingerprints: List[str]
    output_fingerprints: List[str]
    logs: List[GradingLogsHolder]


class NoCacheException(Exception):
    pass


def _check_digests(artifacts_list: List[GradingArtifacts]):
    produced = set()
    for artifacts in artifacts_list:
        for input in artifacts.inputs:
            if input.digest is None:
                continue
            if input.digest.value is not None:
                continue
            if id(input.digest) not in produced:
                raise ValueError('Digests must be produced before being consumed')
        for output in artifacts.outputs:
            if output.digest is None:
                continue
            if output.digest.value is not None:
                continue
            if id(output.digest) in produced:
                raise ValueError('A digest cannot be produced more than once')
            produced.add(id(output.digest))


def _build_artifact_with_digest_list(
    artifacts_list: List[GradingArtifacts],
) -> List[GradingFileOutput]:
    outputs = []
    for artifacts in artifacts_list:
        for output in artifacts.outputs:
            if output.hash and output.digest is None:
                output.digest = DigestHolder()
            if output.digest is None:
                continue
            outputs.append(output)
    return outputs


def _build_digest_list(artifacts_list: List[GradingArtifacts]) -> List[DigestHolder]:
    outputs = _build_artifact_with_digest_list(artifacts_list)
    digests = []
    for output in outputs:
        assert output.digest is not None
        digests.append(output.digest)
    return digests


def _build_fingerprint_list(
    artifacts_list: List[GradingArtifacts], cacher: FileCacher
) -> List[str]:
    fingerprints = []
    for artifacts in artifacts_list:
        for input in artifacts.inputs:
            if input.src is None or not input.hash:
                continue
            if cacher.digest_from_symlink(input.src) is not None:
                continue
            with input.src.open('rb') as f:
                fingerprints.append(digest_cooperatively(f))
    return fingerprints


def _maybe_check_integrity(output: GradingFileOutput, integrity_digest: str):
    if not grading_context.should_check_integrity():
        return
    if not output.hash:
        return
    if output.dest is None or not output.dest.is_symlink() or not output.dest.is_file():
        # Only makes sense if the file EXISTS and IS A SYMLINK pointing to an
        # EXISTING storage file.
        # If the storage file ceases to exist, we can simply evict from the cache.
        return
    if output.digest is None:
        return
    with output.dest.open('rb') as f:
        output_digest = digest_cooperatively(f)
    if output_digest != integrity_digest:
        raise ValueError(
            f'Cache was tampered with, file {output.dest} has changed since it was cached.\nPlease run `rbx clean` to reset the cache.'
        )


def _check_digest_list_integrity(
    artifacts_list: List[GradingArtifacts], integrity_digests: List[Optional[str]]
):
    outputs = _build_artifact_with_digest_list(artifacts_list)
    assert len(outputs) == len(integrity_digests)
    for output, integrity_digest in zip(outputs, integrity_digests):
        assert output.digest is not None
        if integrity_digest is None:
            continue
        _maybe_check_integrity(output, integrity_digest)


def _build_output_fingerprint_list(
    artifacts_list: List[GradingArtifacts],
) -> List[str]:
    fingerprints = []
    for artifacts in artifacts_list:
        for output in artifacts.outputs:
            if output.dest is None or output.intermediate or output.hash:
                continue
            if not output.dest.is_file():
                fingerprints.append('')  # file does not exist
                continue
            with output.dest.open('rb') as f:
                fingerprints.append(digest_cooperatively(f))
    return fingerprints


def _build_logs_list(artifacts_list: List[GradingArtifacts]) -> List[GradingLogsHolder]:
    logs = []
    for artifacts in artifacts_list:
        if artifacts.logs is not None:
            logs.append(artifacts.logs)
    return logs


def _build_cache_fingerprint(
    artifacts_list: List[GradingArtifacts],
    cacher: FileCacher,
) -> CacheFingerprint:
    digests = [digest.value for digest in _build_digest_list(artifacts_list)]
    fingerprints = _build_fingerprint_list(artifacts_list, cacher)
    output_fingerprints = _build_output_fingerprint_list(
        artifacts_list,
    )

    logs = _build_logs_list(artifacts_list)
    return CacheFingerprint(
        digests=digests,
        fingerprints=fingerprints,
        output_fingerprints=output_fingerprints,
        logs=logs,
    )


def _fingerprints_match(
    fingerprint: CacheFingerprint, reference: CacheFingerprint
) -> bool:
    lhs, rhs = fingerprint.fingerprints, reference.fingerprints
    return tuple(lhs) == tuple(rhs)


def _output_fingerprints_match(
    fingerprint: CacheFingerprint, reference: CacheFingerprint
) -> bool:
    lhs, rhs = fingerprint.output_fingerprints, reference.output_fingerprints
    return tuple(lhs) == tuple(rhs)


def _build_cache_input(
    commands: List[str],
    artifact_list: List[GradingArtifacts],
    extra_params: Dict[str, Any],
    cacher: FileCacher,
) -> CacheInput:
    cloned_artifact_list = [
        artifacts.model_copy(deep=True) for artifacts in artifact_list
    ]
    for artifacts in cloned_artifact_list:
        # Clear logs from cache input, since they are not
        # part of the cache key.
        artifacts.logs = None

        for input in artifacts.inputs:
            if input.src is None:
                continue
            inferred_digest = cacher.digest_from_symlink(input.src)
            if inferred_digest is not None:
                # Consume cache from digest instead of file.
                input.digest = DigestHolder(value=inferred_digest)
                input.src = None

        for output in artifacts.outputs:
            if output.hash:
                # Cleanup dest field from hash artifacts
                # since they only their digest value should
                # be tracked by cache.
                output.dest = None

            if output.digest is not None:
                # Cleanup output digest value from cache input,
                # since it is not part of the cache key.
                output.digest.value = None
    return CacheInput(
        commands=commands, artifacts=cloned_artifact_list, extra_params=extra_params
    )


def _build_cache_key(input: CacheInput) -> str:
    with io.BytesIO(input.model_dump_json().encode()) as fobj:
        return digest_cooperatively(fobj)


def _copy_hashed_files(artifact_list: List[GradingArtifacts], cacher: FileCacher):
    for artifact in artifact_list:
        for output in artifact.outputs:
            if not output.hash or output.dest is None:
                continue
            assert output.digest is not None
            if output.optional and output.digest.value is None:
                continue
            assert output.digest.value is not None
            if (
                path_to_symlink := cacher.path_for_symlink(output.digest.value)
            ) is not None:
                # Use a symlink to the file in the persistent cache, if available.
                output.dest.unlink(missing_ok=True)
                output.dest.parent.mkdir(parents=True, exist_ok=True)
                output.dest.symlink_to(path_to_symlink)
            else:
                # Otherwise, copy it.
                with cacher.get_file(output.digest.value) as fobj:
                    with output.dest.open('wb') as f:
                        copyfileobj(fobj, f, maxlen=output.maxlen)
            if output.executable:
                output.dest.chmod(0o755)


def is_artifact_ok(artifact: GradingArtifacts, cacher: FileCacher) -> bool:
    for output in artifact.outputs:
        if output.optional or output.intermediate:
            continue
        if output.digest is not None:
            if output.digest.value is None or not cacher.exists(output.digest.value):
                return False
            return True
        assert output.dest is not None
        file_path: pathlib.Path = artifact.root / output.dest
        if not file_path.is_file():
            return False
        executable = os.access(str(file_path), os.X_OK)
        if executable != output.executable:
            return False
    return True


def are_artifacts_ok(artifacts: List[GradingArtifacts], cacher: FileCacher) -> bool:
    for artifact in artifacts:
        if not is_artifact_ok(artifact, cacher):
            return False
    return True


class DependencyCacheBlock:
    class Break(Exception):
        pass

    def __init__(
        self,
        cache: 'DependencyCache',
        commands: List[str],
        artifact_list: List[GradingArtifacts],
        extra_params: Dict[str, Any],
    ):
        self.cache = cache
        self.commands = commands
        self.artifact_list = artifact_list
        self.extra_params = extra_params
        self._key = None

    def __enter__(self):
        with Profiler('enter_in_cache'):
            if grading_context.is_no_cache():
                return False
            input = _build_cache_input(
                commands=self.commands,
                artifact_list=self.artifact_list,
                extra_params=self.extra_params,
                cacher=self.cache.cacher,
            )
            if VERBOSE:
                console.console.log(f'Cache input is: {input}')
            self._key = _build_cache_key(input)
            if VERBOSE:
                console.console.log(f'Cache key is: {self._key}')
            found = self.cache.find_in_cache(
                self.commands, self.artifact_list, self.extra_params, key=self._key
            )
            return found

    def __exit__(self, exc_type, exc_val, exc_tb):
        with Profiler('exit_in_cache'):
            if grading_context.is_no_cache():
                return True if exc_type is NoCacheException else None
            if exc_type is None:
                self.cache.store_in_cache(
                    self.commands, self.artifact_list, self.extra_params, key=self._key
                )
            if exc_type is NoCacheException:
                return True
            return None


class DependencyCache:
    root: pathlib.Path
    cacher: FileCacher

    def __init__(self, root: pathlib.Path, cacher: FileCacher):
        self.root = root
        self.cacher = cacher
        self.db = SqliteDict(self._cache_name(), autocommit=True)
        tmp_dir = pathlib.Path(tempfile.mkdtemp())
        self.transient_db = SqliteDict(str(tmp_dir / '.cache_db'), autocommit=True)
        atexit.register(lambda: self.db.close())
        atexit.register(lambda: self.transient_db.close())
        atexit.register(lambda: shutil.rmtree(tmp_dir, ignore_errors=True))

    def _cache_name(self) -> str:
        return str(self.root / '.cache_db')

    def get_db(self) -> shelve.Shelf:
        if grading_context.is_transient():
            return self.transient_db
        return self.db

    def _find_in_cache(self, key: str) -> Optional[CacheFingerprint]:
        return self.get_db().get(key)

    def _store_in_cache(self, key: str, fingerprint: CacheFingerprint):
        self.get_db()[key] = fingerprint

    def _evict_from_cache(self, key: str):
        db = self.get_db()
        if key in db:
            del db[key]

    def __call__(
        self,
        commands: List[str],
        artifact_list: List[GradingArtifacts],
        extra_params: Optional[Dict[str, Any]] = None,
    ) -> DependencyCacheBlock:
        _check_digests(artifact_list)
        return DependencyCacheBlock(self, commands, artifact_list, extra_params or {})

    def find_in_cache(
        self,
        commands: List[str],
        artifact_list: List[GradingArtifacts],
        extra_params: Dict[str, Any],
        key: Optional[str] = None,
    ) -> bool:
        input = _build_cache_input(
            commands=commands,
            artifact_list=artifact_list,
            extra_params=extra_params,
            cacher=self.cacher,
        )
        key = key or _build_cache_key(input)

        fingerprint = self._find_in_cache(key)
        if fingerprint is None:
            return False

        reference_fingerprint = _build_cache_fingerprint(
            artifact_list,
            self.cacher,
        )

        if not _fingerprints_match(fingerprint, reference_fingerprint):
            self._evict_from_cache(key)
            return False

        if not _output_fingerprints_match(fingerprint, reference_fingerprint):
            self._evict_from_cache(key)
            return False

        # Check whether existing storage files were not tampered with.
        _check_digest_list_integrity(
            artifact_list,
            fingerprint.digests,
        )
        reference_digests = _build_digest_list(artifact_list)

        # Apply digest changes.
        old_digest_values = [digest for digest in reference_fingerprint.digests]
        for digest, reference_digest in zip(fingerprint.digests, reference_digests):
            reference_digest.value = digest

        if not are_artifacts_ok(artifact_list, self.cacher):
            # Rollback digest changes.
            for old_digest_value, reference_digest in zip(
                old_digest_values, reference_digests
            ):
                reference_digest.value = old_digest_value
            self._evict_from_cache(key)
            return False

        # Copy hashed files to file system.
        _copy_hashed_files(artifact_list, self.cacher)

        # Apply logs changes.
        for logs, reference_logs in zip(fingerprint.logs, reference_fingerprint.logs):
            if logs.run is not None:
                reference_logs.run = logs.run.model_copy(deep=True)
            if logs.interactor_run is not None:
                reference_logs.interactor_run = logs.interactor_run.model_copy(
                    deep=True
                )
            if logs.preprocess is not None:
                reference_logs.preprocess = [
                    log.model_copy(deep=True) for log in logs.preprocess
                ]
            reference_logs.cached = True

        return True

    def store_in_cache(
        self,
        commands: List[str],
        artifact_list: List[GradingArtifacts],
        extra_params: Dict[str, Any],
        key: Optional[str] = None,
    ):
        input = _build_cache_input(
            commands=commands,
            artifact_list=artifact_list,
            extra_params=extra_params,
            cacher=self.cacher,
        )
        key = key or _build_cache_key(input)

        if not are_artifacts_ok(artifact_list, self.cacher):
            return

        reference_fingerprint = _build_cache_fingerprint(
            artifact_list,
            self.cacher,
        )
        self._store_in_cache(
            key,
            reference_fingerprint,
        )



import dataclasses
import pathlib
from abc import ABC, abstractmethod
from typing import Any, Dict, List

import typer

from rbx import console
from rbx.box.statements.builders import (
    StatementBuilderContest,
    StatementCodeLanguage,
)
from rbx.box.statements.schema import Joiner, JoinerType, JoinTexToPDF, StatementType


@dataclasses.dataclass
class StatementJoinerContext:
    languages: List[StatementCodeLanguage]
    params: Joiner
    root: pathlib.Path

    def build_jinja_kwargs(self) -> Dict[str, Any]:
        return {'languages': self.languages}


class StatementJoiner(ABC):
    @abstractmethod
    def name(self) -> JoinerType:
        pass

    @abstractmethod
    def default_params(self) -> Joiner:
        pass

    @abstractmethod
    def input_type(self) -> StatementType:
        pass

    @abstractmethod
    def output_type(self) -> StatementType:
        pass

    @abstractmethod
    def joined_type(self) -> StatementType:
        pass

    @abstractmethod
    def build(
        self,
        input: bytes,
        context: StatementJoinerContext,
        contest: StatementBuilderContest,
        verbose: bool = False,
    ) -> bytes:
        pass


class TeX2PDFJoiner(StatementJoiner):
    def name(self) -> JoinerType:
        return JoinerType.TexToPDF

    def default_params(self) -> Joiner:
        return JoinTexToPDF(type=JoinerType.TexToPDF)

    def input_type(self) -> StatementType:
        return StatementType.TeX

    def output_type(self) -> StatementType:
        return StatementType.PDF

    def joined_type(self) -> StatementType:
        return StatementType.TeX

    def build(
        self,
        input: bytes,
        context: StatementJoinerContext,
        contest: StatementBuilderContest,
        verbose: bool = False,
    ) -> bytes:
        from rbx.box.statements.latex import (
            MAX_PDFLATEX_RUNS,
            Latex,
            decode_latex_output,
            should_rerun,
        )

        latex = Latex(input.decode())
        latex_result = latex.build_pdf(context.root)
        pdf = latex_result.pdf
        logs = decode_latex_output(latex_result.result.stdout)
        runs = 1

        while pdf is not None and should_rerun(logs) and runs < MAX_PDFLATEX_RUNS:
            console.console.print(
                'Re-running pdfLaTeX to get cross-references right...'
            )
            latex_result = latex.build_pdf(context.root)
            pdf = latex_result.pdf
            logs = decode_latex_output(latex_result.result.stdout)
            runs += 1

        if pdf is None:
            console.console.print(f'{logs}')
            console.console.print('[error]PdfLaTeX compilation failed.[/error]')
            raise typer.Exit(1)

        if verbose:
            console.console.print(f'{logs}')

        return pdf


JOINER_LIST: List[StatementJoiner] = [TeX2PDFJoiner()]



import pathlib
import re
from abc import ABC, abstractmethod
from typing import List, Optional, Tuple, Union

import typer

from rbx import console, utils
from rbx.box import cd, package
from rbx.box.formatting import href, ref
from rbx.box.tooling.boca.scraper import BocaRun

PathLike = Union[str, pathlib.Path]


class Expander(ABC):
    def get_remote_path(self, path: pathlib.Path) -> pathlib.Path:
        return package.get_problem_remote_dir() / path

    def cacheable_paths(self, path: pathlib.Path) -> List[pathlib.Path]:
        return []

    def cacheable_globs(self, path: pathlib.Path) -> List[str]:
        return []

    @abstractmethod
    def expand(self, path: pathlib.Path) -> Optional[pathlib.Path]:
        pass


class MainExpander(Expander):
    def expand(self, path: pathlib.Path) -> Optional[pathlib.Path]:
        if str(path) != '@main':
            return None
        sol = package.get_main_solution()
        if sol is None:
            return None
        return sol.path


class BocaExpander(Expander):
    BOCA_REGEX = re.compile(r'\@boca\/(\d+)(?:\-(\d+))?')

    def get_match(self, path_str: str) -> Optional[Tuple[int, int]]:
        match = self.BOCA_REGEX.match(path_str)
        if match is None:
            return None
        run_number = int(match.group(1))
        site_number = int(match.group(2)) if match.group(2) is not None else 1
        return run_number, site_number

    def get_boca_folder(self) -> pathlib.Path:
        return self.get_remote_path(pathlib.Path('boca'))

    def get_boca_path(self, run_number: int, site_number: int) -> pathlib.Path:
        return self.get_boca_folder() / f'{run_number}-{site_number}'

    def cacheable_globs(self, path: pathlib.Path) -> List[str]:
        match = self.get_match(str(path))
        if match is None:
            return []
        run_number, site_number = match
        return [str(self.get_boca_path(run_number, site_number)) + '.*']

    def expand(self, path: pathlib.Path) -> Optional[pathlib.Path]:
        from rbx.box.tooling.boca import scraper as boca_upload

        match = self.get_match(str(path))
        if match is None:
            return None
        run_number, site_number = match

        run = BocaRun.from_run_number(run_number, site_number)
        boca_uploader = boca_upload.get_boca_scraper()
        boca_uploader.login()
        sol_path = boca_uploader.download_run(run, self.get_boca_folder())
        console.console.print(f'Downloaded {href(sol_path)} from BOCA...')
        return sol_path


REGISTERED_EXPANDERS: List['Expander'] = [
    MainExpander(),
    BocaExpander(),
]


def _relative_to_pkg(path: pathlib.Path) -> pathlib.Path:
    return utils.abspath(path).relative_to(pathlib.Path.cwd())


def _try_cacheable_paths(
    path: pathlib.Path, expander: Expander
) -> Optional[pathlib.Path]:
    cached_paths = expander.cacheable_paths(path)
    for cached_path in cached_paths:
        if cached_path.exists():
            return _relative_to_pkg(cached_path)
    return None


def _try_cacheable_globs(
    path: pathlib.Path, expander: Expander
) -> Optional[pathlib.Path]:
    cached_globs = expander.cacheable_globs(path)
    for cached_glob in cached_globs:
        rel_glob = _relative_to_pkg(pathlib.Path(cached_glob))
        globbed = list(pathlib.Path.cwd().glob(str(rel_glob)))
        if not globbed:
            continue
        return _relative_to_pkg(globbed[0])
    return None


def _try_cache(path: pathlib.Path, expander: Expander) -> Optional[pathlib.Path]:
    cached = _try_cacheable_paths(path, expander)
    if cached is not None:
        return cached
    return _try_cacheable_globs(path, expander)


def _expand_path(path: pathlib.Path) -> Optional[pathlib.Path]:
    if not cd.is_problem_package():
        console.console.print(
            f'Skipping expansion of {ref(path)} because we are not in a problem package.'
        )
        raise typer.Exit(1)

    for expander in REGISTERED_EXPANDERS:
        cached = _try_cache(path, expander)
        if cached is not None:
            return cached
        expanded = expander.expand(path)
        if expanded is not None:
            return _relative_to_pkg(expanded)
    return None


def _expand_paths(paths: List[pathlib.Path]) -> List[pathlib.Path]:
    res = []
    for path in paths:
        if not str(path).startswith('@'):
            res.append(path)
            continue
        expanded = _expand_path(path)
        if expanded is None:
            console.console.print(
                f'[warning]Remote solution [item]{path}[/item] could not be expanded. Skipping.[/warning]'
            )
            continue
        res.append(expanded)
    return res


def expand_files(files: List[str]) -> List[pathlib.Path]:
    return _expand_paths([pathlib.Path(file) for file in files])


def expand_file(file: str) -> pathlib.Path:
    res = expand_files([file])
    if len(res) != 1:
        console.console.print(
            f'Could not expand {ref(file)} because it is not a valid expansion.'
        )
        raise typer.Exit(1)
    return res[0]


def is_path_remote(path: pathlib.Path) -> bool:
    remote_dir = package.get_problem_remote_dir()
    return utils.abspath(path).is_relative_to(utils.abspath(remote_dir))



import difflib
import pathlib
from typing import Optional, Type

import yamlfix
import yamlfix.model
from pydantic import BaseModel

from rbx import console
from rbx.box.cd import is_contest_package, is_preset_package, is_problem_package
from rbx.box.contest.schema import Contest
from rbx.box.presets import get_preset_yaml
from rbx.box.presets.schema import Preset
from rbx.box.schema import Package
from rbx.box.stats import find_problem_packages_from_contest
from rbx.utils import uploaded_schema_path


def fix_language_server(path: pathlib.Path, model_cls: Type[BaseModel]) -> bool:
    stream = []
    with path.open('r') as f:
        for line in f:
            if line.strip().startswith('# yaml-language-server:'):
                continue
            stream.append(line)
            if line.startswith('---'):
                stream.append(
                    f'# yaml-language-server: $schema={uploaded_schema_path(model_cls)}\n'
                )
    content = ''.join(stream)
    orig_text = path.read_text()
    path.write_text(content)
    return orig_text != content


def fix_yaml(
    path: pathlib.Path,
    verbose: bool = True,
    print_diff: bool = False,
    model_cls: Optional[Type[BaseModel]] = None,
):
    orig_text = path.read_text()

    # Config to go hand-to-hand with VSCode YAML extension,
    # which we offer first class support to. Unfortunately,
    # YAML extension is not perfect :(
    config = yamlfix.model.YamlfixConfig(
        quote_basic_values=True,
        quote_representation='"',
        comments_min_spaces_from_content=1,
    )
    _, changed = yamlfix.fix_files([str(path)], dry_run=False, config=config)

    # if model_cls is not None:
    #     if fix_language_server(path, model_cls):
    #         changed = True

    if changed and verbose:
        console.console.print(
            f'Formatting [item]{path}[/item].',
        )

    if print_diff and changed:
        unified_diff = difflib.unified_diff(
            orig_text.splitlines(), path.read_text().splitlines()
        )
        console.console.print(
            f'Diff for [item]{path}[/item].\n' + '\n'.join(unified_diff),
        )


def fix_package(root: pathlib.Path = pathlib.Path(), print_diff: bool = False):
    if is_preset_package(root):
        fix_yaml(root / 'preset.rbx.yml', model_cls=Preset, print_diff=print_diff)
        preset = get_preset_yaml(root)
        if preset.problem is not None:
            fix_yaml(
                root / preset.problem / 'problem.rbx.yml',
                model_cls=Package,
                print_diff=print_diff,
            )
        if preset.contest is not None:
            fix_package(root / preset.contest, print_diff=print_diff)
        return

    if is_problem_package(root):
        fix_yaml(root / 'problem.rbx.yml', model_cls=Package, print_diff=print_diff)
    if is_contest_package(root):
        fix_yaml(root / 'contest.rbx.yml', model_cls=Contest, print_diff=print_diff)
        for problem in find_problem_packages_from_contest(root):
            fix_yaml(
                problem / 'problem.rbx.yml',
                model_cls=Package,
                print_diff=print_diff,
            )



"""This module provides a template-rendering function for Jinja2
that overrides Jinja2 defaults to make it work more seamlessly
with Latex.
"""

import pathlib
import re
import typing
from typing import Any, Dict, Optional, Tuple, Union

import jinja2
import jinja2.runtime
import rich.pretty
import typer

from rbx import console

######################################################################
# J2_ARGS
#   Constant was borrowed from Marc Brinkmann's
#   latex repository (mbr/latex on github)
######################################################################
J2_ARGS = {
    'block_start_string': r'\BLOCK{',
    'block_end_string': '}',
    'variable_start_string': r'\VAR{',
    'variable_end_string': '}',
    'comment_start_string': r'\#{',
    'comment_end_string': '}',
    'line_statement_prefix': '%-',
    'line_comment_prefix': '%#',
    'trim_blocks': True,
    'autoescape': False,
}

J2_MD_ARGS = {
    'trim_blocks': True,
    'autoescape': False,
}


######################################################################
# Latex escape regex constants
######################################################################

# Organize all latex escape characters in one list
# (EXCEPT FOR ( "\" ), which is handled separately)
# escaping those which are special characters in
# PERL regular expressions
ESCAPE_CHARS = [
    r'\&',
    '%',
    r'\$',
    '#',
    '_',
    r'\{',
    r'\}',
    '~',
    r'\^',
]

# For each latex escape character, create a regular expression
# that matches all of the following criteria
# 1) one or two characters
# 2) if two characters, the first character is NOT a backslash ( "\" )
# 3) if two characters, the second, if one, the first character
#       is one of the latex escape characters
REGEX_ESCAPE_CHARS = [
    (re.compile(r'(?<!\\)' + i), r'\\' + i.replace('\\', '')) for i in ESCAPE_CHARS
]

# Place escape characters in [] for "match any character" regex
ESCAPE_CHARS_OR = r'[{}\\]'.format(''.join(ESCAPE_CHARS))

# For the back slash, create a regular expression
# that matches all of the following criteria
# 1) one, two, or three characters
# 2) the first character is not a backslash
# 3) the second character is a backslash
# 4) the third character is none of the ESCAPE_CHARS,
#       and is also not a backslash
REGEX_BACKSLASH = re.compile(r'(?<!\\)\\(?!{})'.format(ESCAPE_CHARS_OR))


######################################################################
# Declare module functions
######################################################################
def escape_latex_str_if_str(value):
    """Escape a latex string"""
    if not isinstance(value, str):
        return value
    for regex, replace_text in REGEX_ESCAPE_CHARS:
        value = re.sub(regex, replace_text, value)
    value = re.sub(REGEX_BACKSLASH, r'\\textbackslash{}', value)
    return value


def _process_zeroes(value: int) -> Tuple[int, int, int]:
    cnt = 0

    acc = value
    while acc >= 10:
        acc //= 10
        cnt += 1
    return acc, cnt, value - acc * 10**cnt


def scientific_notation(
    value: Union[int, jinja2.Undefined], zeroes: int = 2
) -> Union[str, jinja2.Undefined]:
    if jinja2.is_undefined(value):
        return typing.cast(jinja2.Undefined, value)
    assert isinstance(value, int)
    assert zeroes >= 1
    if value == 0:
        return '0'
    if value < 0:
        return f'-{scientific_notation(-value, zeroes=zeroes)}'

    mult, exp, rest = _process_zeroes(value)
    if exp < zeroes:
        return str(value)
    res = '10' if exp == 1 else f'10^{{{exp}}}'
    if rest > 0 and len(str(rest)) + 1 >= len(str(value)):
        # Should not convert numbers like 532 to 5*10^2 + 32.
        return str(value)
    if mult > 1:
        res = f'{mult} \\times {res}'
    if rest > 0:
        res = f'{res} + {rest}'
    return res


def path_parent(path: pathlib.Path) -> pathlib.Path:
    return path.parent


def path_stem(path: pathlib.Path) -> str:
    return path.stem


@jinja2.pass_context
def test_var_truthy(ctx: jinja2.runtime.Context, value: Any):
    if isinstance(value, jinja2.Undefined):
        return False
    if value is None:
        return False
    return bool(value)


@jinja2.pass_context
def test_var_falsy(ctx: jinja2.runtime.Context, value: Any):
    return not test_var_truthy(ctx, value)


@jinja2.pass_context
def test_var_null(ctx: jinja2.runtime.Context, value: Any):
    if isinstance(value, jinja2.Undefined):
        return True
    if value is None:
        return True
    return False


@jinja2.pass_context
def test_var_nonnull(ctx: jinja2.runtime.Context, value: Any):
    return not test_var_null(ctx, value)


######################################################################
# Declare module functions
######################################################################


class StrictChainableUndefined(jinja2.StrictUndefined):
    def __getattr__(self, name: str) -> 'StrictChainableUndefined':
        # Raise AttributeError on requests for names that appear to be unimplemented
        # dunder methods to avoid confusing Python with truthy non-method objects that
        # do not implement the protocol being probed for. e.g., copy.copy(Undefined())
        # fails spectacularly if getattr(Undefined(), '__setstate__') returns an
        # Undefined object instead of raising AttributeError to signal that it does not
        # support that style of object initialization.
        if name[:2] == '__' and name[-2:] == '__':
            raise AttributeError(name)

        return self

    def __getitem__(self, _name: str) -> 'StrictChainableUndefined':  # type: ignore[override]
        return self


class VarWrapperUndefinedError(jinja2.UndefinedError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def vars(self) -> Dict[str, Any]:
        return {}


class JinjaDictWrapper(dict):
    def __init__(
        self,
        *args,
        key='dict object',
        prefix='',
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.key = key
        self.prefix = prefix
        self.ancestor_d: Optional[Dict[str, Any]] = None

        slf = self

        class _AccessError(VarWrapperUndefinedError):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)

            def vars(self) -> Dict[str, Any]:
                return slf.ancestor_d or slf

        self.exc = _AccessError

    @classmethod
    def from_dict(cls, d: Dict[str, Any], wrapper_key: str) -> 'JinjaDictWrapper':
        res = cls(key=wrapper_key)
        for key, value in d.items():
            splits = key.split('.')
            prefix = ''
            acc = res
            for split in splits[:-1]:
                prefix = f'{prefix}.{split}'.strip('.')
                if split not in acc or not isinstance(acc[split], dict):
                    acc[split] = JinjaDictWrapper(key=wrapper_key, prefix=prefix)
                    acc[split].ancestor_d = res
                acc = acc[split]
            acc[splits[-1]] = value
        return res

    def __getitem__(self, key: str) -> Any:
        try:
            return super().__getitem__(key)
        except KeyError:
            final_key = f'{self.prefix}.{key}'.strip('.')
            return StrictChainableUndefined(
                hint=f'"{final_key}" was not found in "{self.key}"',
                exc=self.exc,
            )


def add_builtin_filters(j2_env: jinja2.Environment):
    j2_env.filters['escape'] = escape_latex_str_if_str
    j2_env.filters['sci'] = scientific_notation
    j2_env.filters['parent'] = path_parent
    j2_env.filters['stem'] = path_stem


def add_builtin_tests(j2_env: jinja2.Environment):
    j2_env.tests['truthy'] = test_var_truthy
    j2_env.tests['falsy'] = test_var_falsy
    j2_env.tests['null'] = test_var_null
    j2_env.tests['nonnull'] = test_var_nonnull


def _handle_rendering_undefined(
    err: jinja2.UndefinedError,
) -> str:
    console.console.print('[error]Error while rendering Jinja2 template:', end=' ')
    console.console.print(err)
    console.console.print(
        '[warning]This usually happens when accessing an undefined variable.[/warning]'
    )
    if isinstance(err, VarWrapperUndefinedError):
        vars = err.vars()
        if vars:
            console.console.print('[warning]Defined variables are[/warning] ', end='')
            console.console.print(rich.pretty.Pretty(vars))
    raise typer.Abort() from err


def render_latex_template(path_templates, template_filename, template_vars=None) -> str:
    """Render a latex template, filling in its template variables

    :param path_templates: the path to the template directory
    :param template_filename: the name, rooted at the path_template_directory,
        of the desired template for rendering
    :param template_vars: dictionary of key:val for jinja2 variables
        defaults to None for case when no values need to be passed
    """
    var_dict = template_vars if template_vars else {}
    j2_env = jinja2.Environment(
        loader=jinja2.FileSystemLoader(path_templates),
        **J2_ARGS,
        undefined=StrictChainableUndefined,
    )
    add_builtin_filters(j2_env)
    add_builtin_tests(j2_env)
    template = j2_env.get_template(template_filename)
    try:
        return template.render(**var_dict)  # type: ignore
    except jinja2.UndefinedError as err:
        _handle_rendering_undefined(err)
        raise


def render_latex_template_blocks(
    path_templates, template_filename, template_vars=None
) -> Dict[str, str]:
    """Render a latex template, filling in its template variables

    :param path_templates: the path to the template directory
    :param template_filename: the name, rooted at the path_template_directory,
        of the desired template for rendering
    :param template_vars: dictionary of key:val for jinja2 variables
        defaults to None for case when no values need to be passed
    """
    var_dict = template_vars if template_vars else {}
    j2_env = jinja2.Environment(
        loader=jinja2.FileSystemLoader(path_templates),
        **J2_ARGS,
        undefined=StrictChainableUndefined,
    )
    add_builtin_filters(j2_env)
    add_builtin_tests(j2_env)
    template = j2_env.get_template(template_filename)
    ctx = template.new_context(var_dict)  # type: ignore
    try:
        return {key: ''.join(value(ctx)) for key, value in template.blocks.items()}
    except jinja2.UndefinedError as err:
        _handle_rendering_undefined(err)
        raise


def render_markdown_template_blocks(
    path_templates, template_filename, template_vars=None
) -> Dict[str, str]:
    """Render a markdown template, filling in its template variables

    :param path_templates: the path to the template directory
    :param template_filename: the name, rooted at the path_template_directory,
        of the desired template for rendering
    :param template_vars: dictionary of key:val for jinja2 variables
        defaults to None for case when no values need to be passed
    """
    var_dict = template_vars if template_vars else {}
    j2_env = jinja2.Environment(
        loader=jinja2.FileSystemLoader(path_templates),
        **J2_MD_ARGS,  # type: ignore
        undefined=StrictChainableUndefined,
    )
    add_builtin_filters(j2_env)
    add_builtin_tests(j2_env)
    template = j2_env.get_template(template_filename)
    ctx = template.new_context(var_dict)  # type: ignore
    try:
        return {key: ''.join(value(ctx)) for key, value in template.blocks.items()}
    except jinja2.UndefinedError as err:
        _handle_rendering_undefined(err)
        raise



import pathlib
import shutil
import time
from shutil import rmtree
from typing import List, Optional

import async_lru
import syncer
import typer
from pydantic import BaseModel

from rbx import console, utils
from rbx.box import checkers, generators, package, tasks, validators
from rbx.box.code import SanitizationLevel, compile_item
from rbx.box.generators import (
    GenerationMetadata,
    expand_generator_call,
    generate_standalone,
)
from rbx.box.schema import CodeItem, GeneratorCall, Stress, TaskType, Testcase
from rbx.box.solutions import compile_solutions, get_outcome_style_verdict
from rbx.box.stressing import finder_parser
from rbx.grading.steps import (
    Evaluation,
    Outcome,
)
from rbx.utils import StatusProgress


class StressFinding(BaseModel):
    generator: GeneratorCall


class StressReport(BaseModel):
    findings: List[StressFinding] = []
    executed: int = 0


def _compile_finder(finder: CodeItem) -> str:
    try:
        digest = compile_item(finder)
    except Exception as e:
        console.console.print(
            f'[error]Failed compiling checker [item]{finder.path}[/item][/error]'
        )
        raise typer.Exit(1) from e
    return digest


async def run_stress(
    timeoutInSeconds: int,
    name: Optional[str] = None,
    finder: Optional[str] = None,
    generator_call: Optional[str] = None,
    findingsLimit: int = 1,
    verbose: bool = False,
    progress: Optional[StatusProgress] = None,
    sanitized: bool = False,
    print_descriptors: bool = False,
) -> StressReport:
    pkg = package.find_problem_package_or_die()

    if finder:
        if generator_call is None:
            console.console.print(
                '[error]Generator arguments are required for stress testing. Specify them through the [item]-g[/item] flag.[/error]'
            )
            raise typer.Exit(1)
        generator = generators.get_call_from_string(generator_call)
        stress = Stress(
            name=f'{pathlib.Path(generator.name).stem}',
            generator=generator,
            finder=finder,
        )
    else:
        if name is None:
            console.console.print(
                '[error]Invalid stress test paramaters. Either provide a stress test name, or provide a finder expression (-f) and generator arguments (-g).[/error]'
            )
            raise typer.Exit(1)
        stress = package.get_stress(name)

    call = stress.generator
    generator = package.get_generator(call.name)

    if progress:
        progress.update('Compiling generator...')
    try:
        generator_digest = compile_item(generator, sanitized=SanitizationLevel.PREFER)
    except:
        console.console.print(
            f'[error]Failed compiling generator [item]{generator.name}[/item].[/error]'
        )
        raise

    # Finder expression parser
    parsed_finder = finder_parser.parse(stress.finder)

    solutions = finder_parser.get_all_solution_items(parsed_finder)
    finders = finder_parser.get_all_checker_items(parsed_finder)
    needs_expected_output = finder_parser.needs_expected_output(parsed_finder)

    solution_indices = {str(solution.path): i for i, solution in enumerate(solutions)}
    solutions_digest = compile_solutions(
        tracked_solutions=set(str(solution.path) for solution in solutions),
        sanitized=sanitized,
        progress=progress,
    )
    if progress:
        progress.update('Compiling finders...')
    finders_digest = {str(finder.path): _compile_finder(finder) for finder in finders}

    interactor_digest = None
    if pkg.type == TaskType.COMMUNICATION:
        interactor_digest = checkers.compile_interactor(progress=progress)

    if progress:
        progress.update('Compiling validator...')
    compiled_validator = validators.compile_main_validator()

    # Erase old stress directory
    runs_dir = package.get_problem_runs_dir()
    stress_dir = runs_dir / '.stress'
    rmtree(str(stress_dir), ignore_errors=True)
    stress_dir.mkdir(parents=True, exist_ok=True)
    empty_path = runs_dir / '.stress' / '.empty'
    empty_path.write_text('')

    startTime = time.monotonic()

    executed = 0
    findings = []

    while len(findings) < findingsLimit:
        if time.monotonic() - startTime > timeoutInSeconds:
            break

        if print_descriptors:
            utils.print_open_fd_count()

        executed += 1

        if progress:
            seconds = timeoutInSeconds - int(time.monotonic() - startTime)
            progress.update(
                f'Stress testing: found [item]{len(findings)}[/item] tests, '
                f'executed [item]{executed}[/item], '
                f'[item]{seconds}[/item] second(s) remaining...'
            )

        input_path = runs_dir / '.stress' / 'input'
        input_path.parent.mkdir(parents=True, exist_ok=True)

        expanded_generator_call = expand_generator_call(stress.generator)
        await generate_standalone(
            GenerationMetadata(
                generator_call=expanded_generator_call,
                copied_to=Testcase(inputPath=input_path),
            ),
            generator_digest=generator_digest,
            validator_digest=compiled_validator[1]
            if compiled_validator is not None
            else None,
        )

        @async_lru.alru_cache(maxsize=None)
        async def run_solution_fn(
            solution: str,
            checker_digest: Optional[str] = None,
            input_path=input_path,
            output_path: Optional[pathlib.Path] = None,
        ) -> Evaluation:
            index = solution_indices[solution]
            sol = solutions[index]
            return await tasks.run_solution_on_testcase(
                solutions[index],
                compiled_digest=solutions_digest[sol.path],
                checker_digest=checker_digest,
                interactor_digest=interactor_digest,
                testcase=Testcase(inputPath=input_path, outputPath=output_path),
                output_dir=input_path.parent,
                filestem=f'{index}',
                is_stress=True,
            )

        # Get main solution output.
        expected_output_path = empty_path
        if needs_expected_output:
            eval = await run_solution_fn(str(solutions[0].path))
            if eval.result.outcome != Outcome.ACCEPTED:
                console.console.print(
                    '[error]Error while generating main solution output.[/error]'
                )
                console.console.print(f'Input written at [item]{input_path}[/item]')
                console.console.print(
                    f'Output written at [item]{eval.log.stdout_absolute_path}[/item]'
                )
                console.console.print(
                    f'Stderr written at [item]{eval.log.stderr_absolute_path}[/item]'
                )
                console.console.print()
                console.console.print(
                    "[warning]If you don't want reference outputs to be generated for the tests, you should "
                    "use the two-way modifier in your finder expression (':2')."
                )
                raise typer.Exit(1)
            if eval.log.stdout_absolute_path is not None:
                expected_output_path = input_path.with_suffix('.ans')
                shutil.copyfile(eval.log.stdout_absolute_path, expected_output_path)
            else:
                expected_output_path = None

        @async_lru.alru_cache(maxsize=None)
        async def run_solution_and_checker_fn(
            call: finder_parser.FinderCall,
            expected_output_path=expected_output_path,
        ) -> finder_parser.FinderResult:
            async def run_fn() -> Evaluation:
                solution = call.solution
                checker = call.checker

                checker_digest = (
                    finders_digest[checker.path] if checker is not None else None
                )
                return await run_solution_fn(
                    solution,
                    checker_digest=checker_digest,
                    output_path=expected_output_path,
                )

            eval = await run_fn()

            return finder_parser.FinderResult(
                solution=call.solution,
                outcome=eval.result.outcome,
                checker=call.checker,
                solution_log=eval.log,
                checker_result=eval.result,
            )

        @syncer.sync
        async def run_fn(*args, **kwargs):
            # Wrap the runner in a syncer.sync to make it work with the finder parser.
            return await run_solution_and_checker_fn(*args, **kwargs)

        runner = finder_parser.FinderTreeRunner(runner=run_fn)
        finder_outcome: finder_parser.FinderOutcome = runner.transform(parsed_finder)

        internal_error_results = [
            result
            for result in finder_outcome.results
            if result.outcome == Outcome.INTERNAL_ERROR
        ]

        if internal_error_results:
            console.console.print(
                f'[error]Checkers failed during stress test [item]{stress.name}[/item] with args [info]{expanded_generator_call.name} {expanded_generator_call.args}[/info][/error]'
            )
            for internal_error_result in internal_error_results:
                assert internal_error_result.checker is not None
                assert internal_error_result.checker_result is not None
                internal_error_checker_name = internal_error_result.checker.path
                console.console.print(
                    f'[warning]Checker [item]{internal_error_checker_name}[/item] failed with message:'
                )
                console.console.print(internal_error_result.checker_result.message)
            raise typer.Exit(1)

        if not finder_outcome.truth_value:
            continue

        findings_dir = stress_dir / 'findings'
        findings_dir.mkdir(parents=True, exist_ok=True)
        finding_index = len(findings)

        finding_path = findings_dir / f'{finding_index}.in'
        finding_path.write_bytes(input_path.read_bytes())

        if progress:
            console.console.print(
                f'[error]FINDING[/error] Generator args are "[status]{expanded_generator_call.name} {expanded_generator_call.args}[/status]"'
            )
            seen_finder_results = set()
            for finder_result in finder_outcome.results:
                style = get_outcome_style_verdict(finder_result.outcome)
                finder_result_key = (finder_result.solution, finder_result.checker)
                if finder_result_key in seen_finder_results:
                    continue
                seen_finder_results.add(finder_result_key)
                finder_result_report_line = f'{finder_result.solution} = [{style}]{finder_result.outcome.name}[/{style}]'
                if finder_result.checker is not None:
                    finder_result_report_line += (
                        f' [item]ON[/item] {finder_result.checker.path}'
                    )
                console.console.print(finder_result_report_line)

        findings.append(
            StressFinding(
                generator=expanded_generator_call,
            )
        )

        # Be cooperative.
        time.sleep(0.001)

    return StressReport(findings=findings, executed=executed)


def print_stress_report(report: StressReport):
    console.console.rule('Stress test report', style='status')
    console.console.print(f'Executed [item]{report.executed}[/item] tests.')
    if not report.findings:
        console.console.print('No stress test findings.')
        return
    console.console.print(f'Found [item]{len(report.findings)}[/item] testcases.')

    findings_dir = package.get_problem_runs_dir() / '.stress' / 'findings'
    console.console.print(f'Findings: {utils.abspath(findings_dir)}')
    console.console.print()

    for i, finding in enumerate(report.findings):
        console.console.print(f'[error]Finding {i + 1}[/error]')
        console.console.print(
            f'Generator: [status]{finding.generator.name} {finding.generator.args}[/status]'
        )
        console.console.print()



import dataclasses
import pathlib
import shutil
import tempfile
from contextlib import contextmanager
from typing import Awaitable, Callable, List, Optional

from rbx import utils
from rbx.box import package
from rbx.box.setter_config import RepeatsConfig, get_setter_config
from rbx.grading.steps import Evaluation, Outcome


def _both_accepted(eval_a: Evaluation, eval_b: Evaluation) -> bool:
    return (
        eval_a.result.outcome == Outcome.ACCEPTED
        and eval_b.result.outcome == Outcome.ACCEPTED
    )


def _any_tle(eval_a: Evaluation, eval_b: Evaluation) -> bool:
    return eval_a.result.outcome.is_slow() or eval_b.result.outcome.is_slow()


def _get_faster(eval_a: Evaluation, eval_b: Evaluation) -> Evaluation:
    if eval_a.log.time is None:
        return eval_b
    if eval_b.log.time is None:
        return eval_a
    if eval_a.log.time < eval_b.log.time:
        return eval_a
    return eval_b


def _merge_evaluations(eval_a: Evaluation, eval_b: Evaluation) -> Evaluation:
    if _both_accepted(eval_a, eval_b) or _any_tle(eval_a, eval_b):
        return _get_faster(eval_a, eval_b)
    if eval_a.result.outcome != Outcome.ACCEPTED:
        return eval_a
    if eval_b.result.outcome != Outcome.ACCEPTED:
        return eval_b
    return _get_faster(eval_a, eval_b)


@contextmanager
def _temp_retry_dir():
    """Create a temporary directory for retry artifacts."""
    temp_dir = tempfile.mkdtemp(prefix='rbx_retry_')
    try:
        yield pathlib.Path(temp_dir)
    finally:
        shutil.rmtree(temp_dir, ignore_errors=True)


@dataclasses.dataclass
class FileToRecover:
    from_path: pathlib.Path
    to_path: pathlib.Path


def _move_to_temp_dir(path: pathlib.Path, temp_dir: pathlib.Path) -> FileToRecover:
    problem_path = package.find_problem()
    path = utils.abspath(path)
    temp_dir = utils.abspath(temp_dir)
    relative = path.relative_to(problem_path)

    temp_path = temp_dir / relative
    temp_path.parent.mkdir(parents=True, exist_ok=True)
    shutil.move(path, temp_path)
    return FileToRecover(temp_path, path)


def _move_logs_to_temp_dir(
    eval: Evaluation, temp_dir: pathlib.Path
) -> List[FileToRecover]:
    recover = []
    if (
        eval.log.stdout_absolute_path is not None
        and eval.log.stdout_absolute_path.exists()
    ):
        recover.append(_move_to_temp_dir(eval.log.stdout_absolute_path, temp_dir))
    if (
        eval.log.stderr_absolute_path is not None
        and eval.log.stderr_absolute_path.exists()
    ):
        recover.append(_move_to_temp_dir(eval.log.stderr_absolute_path, temp_dir))
    if eval.log.log_absolute_path is not None and eval.log.log_absolute_path.exists():
        recover.append(_move_to_temp_dir(eval.log.log_absolute_path, temp_dir))
    if eval.log.eval_absolute_path is not None and eval.log.eval_absolute_path.exists():
        recover.append(_move_to_temp_dir(eval.log.eval_absolute_path, temp_dir))
    return recover


def get_retrier_config(nruns: int = 0) -> RepeatsConfig:
    if nruns == 0:
        return get_setter_config().repeats
    repeats = get_setter_config().repeats.model_copy(deep=True)
    repeats.reps = nruns
    return repeats


class Retrier:
    def __init__(self, config: Optional[RepeatsConfig] = None, is_stress: bool = False):
        self.config = config or get_setter_config().repeats
        self.is_stress = is_stress

        self.reset()

    def reset(self):
        self.reps = self.config.reps - 1
        self.retries = self.config.retries
        self.retries_for_stress = self.config.retries_for_stress
        self.retry_index = 0

    async def repeat(
        self,
        func: Callable[[int], Awaitable[Evaluation]],
    ) -> Evaluation:
        self.retry_index += 1
        eval = await func(self.retry_index)
        if self.should_repeat(eval):
            with _temp_retry_dir() as temp_dir:
                # Move files to temp dir to open run for repeat.
                recover = _move_logs_to_temp_dir(eval, temp_dir)
                # Actually repeat and choose the best evaluation.
                next_eval = await self.repeat(func)
                chosen_eval = _merge_evaluations(eval, next_eval)

                if id(chosen_eval) == id(eval):
                    # Must recover originally moved files.
                    for file in recover:
                        file.to_path.parent.mkdir(parents=True, exist_ok=True)
                        shutil.move(file.from_path, file.to_path)
                return chosen_eval
        return eval

    def should_repeat(self, eval: Evaluation) -> bool:
        if self.is_stress:
            if eval.result.outcome.is_slow() and self.retries_for_stress > 0:
                self.retries_for_stress -= 1
                return True
        if eval.result.outcome.is_slow() and self.retries > 0:
            self.retries -= 1
            return True
        if self.reps > 0:
            self.reps -= 1
            return True
        return False



import pathlib
from typing import Optional

import aiofiles
from textual import work
from textual.app import ComposeResult
from textual.reactive import reactive
from textual.widget import Widget
from textual.widgets import Log

BATCH_SIZE = 1024


class FileLog(Widget, can_focus=False):
    DEFAULT_CSS = """
    FileLog {
        border: solid $accent;
        height: 1fr;
        width: 1fr;
    }
    """

    path: reactive[Optional[pathlib.Path]] = reactive(None)

    def compose(self) -> ComposeResult:
        yield Log()

    def on_mount(self):
        self.query_one(Log).auto_scroll = False
        self.query_one(Log).can_focus = False

    @work(exclusive=True)
    async def _load_file(self, path: pathlib.Path):
        log = self.query_one(Log)
        log.clear()
        path_str = str(path.relative_to(pathlib.Path.cwd()))
        self.border_subtitle = f'{path_str} (loading...)'

        async with aiofiles.open(path, 'r') as f:
            batch = []
            async for line in f:
                batch.append(line)
                if len(batch) >= BATCH_SIZE:
                    log.write(''.join(batch))
                    batch = []

            if batch:
                log.write(''.join(batch))

        self.border_subtitle = path_str

    async def watch_path(self, path: Optional[pathlib.Path]):
        log = self.query_one(Log)
        log.clear()

        if path is None:
            self.border_subtitle = '(no file selected)'
            return

        if not path.is_file():
            path_str = str(path.relative_to(pathlib.Path.cwd()))
            self.border_subtitle = f'{path_str} (does not exist)'
            return

        self._load_file(path)



import pathlib
from typing import List, Optional, Set

from pydantic import BaseModel

from rbx import console
from rbx.box import checkers, package, validators
from rbx.box.schema import (
    CheckerTest,
    CodeItem,
    ExpectedOutcome,
    Testcase,
    ValidatorOutcome,
    ValidatorTest,
)
from rbx.grading.steps import Outcome
from rbx.utils import StatusProgress


class ValidatorTestEntry(BaseModel):
    input: pathlib.Path
    outcome: ValidatorOutcome
    validator: Optional[CodeItem]


class CheckerTestEntry(BaseModel):
    input: Optional[pathlib.Path] = None
    output: Optional[pathlib.Path] = None
    answer: Optional[pathlib.Path] = None
    outcome: ExpectedOutcome

    def running_tests_formatted_string(self) -> str:
        res = []
        if self.input:
            res.append(f'[item]{self.input}[/item]')
        if self.output:
            res.append(f'[item]{self.output}[/item]')
        if self.answer:
            res.append(f'[item]{self.answer}[/item]')
        return ', '.join(res)


def extract_validator_test_entries(
    tests: List[ValidatorTest],
) -> List[ValidatorTestEntry]:
    res: List[ValidatorTestEntry] = []
    for test in tests:
        for input in sorted(pathlib.Path().glob(str(test.glob))):
            if not input.is_file():
                continue
            res.append(
                ValidatorTestEntry(
                    input=input, outcome=test.outcome, validator=test.validator
                )
            )
    return sorted(res, key=lambda x: x.input.name)


def extract_checker_test_entries(tests: List[CheckerTest]) -> List[CheckerTestEntry]:
    res: List[CheckerTestEntry] = []
    seen: Set[pathlib.Path] = set()
    for test in tests:
        for file in sorted(pathlib.Path().glob(str(test.glob))):
            if not file.is_file():
                continue
            if file.suffix not in ['.in', '.out', '.ans']:
                continue
            basefile = file.with_suffix('')
            if basefile in seen:
                continue
            seen.add(basefile)
            input = basefile.with_suffix('.in')
            output = basefile.with_suffix('.out')
            answer = basefile.with_suffix('.ans')
            res.append(
                CheckerTestEntry(
                    input=input if input.is_file() else None,
                    output=output if output.is_file() else None,
                    answer=answer if answer.is_file() else None,
                    outcome=test.outcome,
                )
            )
    return res


def _get_validator_for_test(test: ValidatorTestEntry) -> Optional[CodeItem]:
    pkg = package.find_problem_package_or_die()
    if test.validator is not None:
        return test.validator
    return pkg.validator


async def run_validator_unit_tests(progress: StatusProgress):
    pkg = package.find_problem_package_or_die()

    entries = extract_validator_test_entries(pkg.unitTests.validator)

    vals: List[CodeItem] = []
    for test in entries:
        val = _get_validator_for_test(test)
        if val is not None:
            vals.append(val)

    console.console.rule('Validator tests', style='info')
    if not entries:
        console.console.print('No validator unit tests found.')
        return

    compiled_validators = validators.compile_validators(vals, progress=progress)

    if progress:
        progress.update('Running validator unit tests...')

    for i, test in enumerate(entries):
        val = _get_validator_for_test(test)
        if val is None:
            console.console.print(
                f'[warning]No validator found for test [item]#{i + 1}[/item], skipping.[/warning]'
            )
            continue

        compiled_digest = compiled_validators[str(val.path)]
        info = await validators.validate_one_off(
            test.input,
            val,
            compiled_digest,
        )

        is_valid = test.outcome == ValidatorOutcome.VALID

        markup = (
            '[success]OK[/success]' if info.ok == is_valid else '[error]FAIL[/error]'
        )

        console.console.print(
            f'{markup} Unit test [item]#{i + 1}[/item] for [item]{test.input}[/item]'
        )
        console.console.print(f'  [status]Expected[/status] {test.outcome.value}')
        if info.ok != is_valid:
            if info.ok:
                console.console.print('  [status]Actual[/status] VALID')
            else:
                console.console.print('  [status]Actual[/status] INVALID')

        if info.message:
            console.console.print(f'  [status]Message[/status] {info.message}')


async def run_checker_unit_tests(progress: StatusProgress):
    pkg = package.find_problem_package_or_die()

    if not package.get_checker():
        console.console.print(
            '[warning]No checker found, skipping checker unit tests.[/warning]'
        )
        return

    console.console.rule('Checker tests', style='info')

    entries = extract_checker_test_entries(pkg.unitTests.checker)
    if not entries:
        console.console.print('No checker unit tests found.')
        return

    compiled_digest = checkers.compile_checker(progress=progress)

    if progress:
        progress.update('Running checker unit tests...')

    empty_file = package.get_empty_sentinel_path()

    for i, test in enumerate(entries):
        result = await checkers.check(
            compiled_digest,
            run_log=None,
            testcase=Testcase(
                inputPath=test.input or empty_file,
                outputPath=test.answer or empty_file,
            ),
            program_output=test.output or empty_file,
            skip_run_log=True,
        )

        if test.answer is not None:
            ans_result = await checkers.check(
                compiled_digest,
                run_log=None,
                testcase=Testcase(
                    inputPath=test.input or empty_file,
                    outputPath=test.answer,
                ),
                program_output=test.answer,
                skip_run_log=True,
            )

            if ans_result.outcome != Outcome.ACCEPTED:
                console.console.print(
                    f'[error]FAIL[/error] Unit test [item]#{i + 1}[/item] ({test.running_tests_formatted_string()})'
                )
                console.console.print(
                    '[error]Error validating the [item].ans[/item] file.'
                )
                console.console.print(
                    '[error]While checking your [item].ans[/item] against itself, the checker returned the following error:[/error]'
                )
                console.console.print(
                    f'  [status]Verdict[/status] {ans_result.outcome.name}'
                )
                console.console.print(
                    f'  [status]Message[/status] {ans_result.message}'
                )
                console.console.print(
                    '[error]Please fix your [item].ans[/item] file and try again, or double-check that your checker is correct.[/error]'
                )
                continue

        markup = (
            '[success]OK[/success]'
            if test.outcome.match(result.outcome)
            else '[error]FAIL[/error]'
        )

        console.console.print(
            f'{markup} Unit test [item]#{i + 1}[/item] ({test.running_tests_formatted_string()})'
        )
        console.console.print(f'  [status]Expected[/status] {test.outcome.name}')

        if not test.outcome.match(result.outcome):
            console.console.print(f'  [status]Actual[/status] {result.outcome.name}')
        if result.message:
            console.console.print(f'  [status]Message[/status] {result.message}')


async def run_unit_tests(progress: StatusProgress):
    await run_validator_unit_tests(progress)
    await run_checker_unit_tests(progress)



import pathlib
import tempfile
from typing import Optional, Type

import syncer
import typer

from rbx import annotations, console
from rbx.box import cd, environment, limits_info, package
from rbx.box.contest import build_contest_statements, contest_package
from rbx.box.packaging.packager import (
    BaseContestPackager,
    BasePackager,
    BuiltContestStatement,
    BuiltProblemPackage,
    ContestZipper,
    run_packager,
)

app = typer.Typer(no_args_is_help=True, cls=annotations.AliasGroup)


async def run_contest_packager(
    contest_packager_cls: Type[BaseContestPackager],
    packager_cls: Type[BasePackager],
    verification: environment.VerificationParam,
    **kwargs,
):
    contest = contest_package.find_contest_package_or_die()

    if limits_info.get_saved_limits_profile(contest_packager_cls.name()) is not None:
        console.console.print(
            f'[warning]Using saved limits profile for [item]{contest_packager_cls.name()}[/item].[/warning]'
        )

    # Build problem-level packages.
    built_packages = []
    for problem in contest.problems:
        console.console.print(
            f'Processing problem [item]{problem.short_name}[/item]...'
        )
        with cd.new_package_cd(problem.get_path()):
            package.clear_package_cache()
            package_path = await run_packager(
                packager_cls, verification=verification, **kwargs
            )
            built_packages.append(
                BuiltProblemPackage(
                    path=problem.get_path() / package_path,
                    package=package.find_problem_package_or_die(),
                    problem=problem,
                )
            )

    # Build statements.
    packager = contest_packager_cls(**kwargs)
    statement_types = packager.statement_types()
    built_statements = []

    with limits_info.use_profile(contest_packager_cls.name()):
        for statement_type in statement_types:
            languages = packager.languages()
            for language in languages:
                statement = packager.get_statement_for_language(language)
                statement_path = build_contest_statements.build_statement(
                    statement, contest, statement_type
                )
                built_statements.append(
                    BuiltContestStatement(statement, statement_path, statement_type)
                )

    console.console.print(f'Packaging contest for [item]{packager.name()}[/item]...')

    # Build contest-level package.
    with tempfile.TemporaryDirectory() as td, limits_info.use_profile(
        contest_packager_cls.name()
    ):
        result_path = packager.package(
            built_packages, pathlib.Path('build'), pathlib.Path(td), built_statements
        )

    console.console.print(
        f'[success]Created contest package for [item]{packager.name()}[/item] at [item]{result_path}[/item]![/success]'
    )


@app.command('polygon', help='Build a contest package for Polygon.')
@contest_package.within_contest
@syncer.sync
async def polygon(
    verification: environment.VerificationParam,
    language: Optional[str] = typer.Option(
        None,
        '--language',
        '-l',
        help='If set, will use the given language as the main language.',
    ),
):
    from rbx.box.packaging.polygon.packager import (
        PolygonContestPackager,
        PolygonPackager,
    )

    await run_contest_packager(
        PolygonContestPackager,
        PolygonPackager,
        verification=verification,
        main_language=language,
    )


@app.command('boca', help='Build a contest package for BOCA.')
@contest_package.within_contest
@syncer.sync
async def boca(
    verification: environment.VerificationParam,
):
    from rbx.box.packaging.boca.packager import BocaPackager

    class BocaContestPackager(ContestZipper):
        def __init__(self, **kwargs):
            super().__init__('boca-contest', zip_inner=True, **kwargs)

        def name(self) -> str:
            return 'boca'

    await run_contest_packager(
        BocaContestPackager, BocaPackager, verification=verification
    )


@app.command('pkg', help='Build a contest package for PKG.')
@contest_package.within_contest
@syncer.sync
async def pkg(
    verification: environment.VerificationParam,
):
    from rbx.box.packaging.pkg.packager import PkgContestPackager, PkgPackager

    await run_contest_packager(
        PkgContestPackager, PkgPackager, verification=verification
    )



from typing import List, Optional

from textual.app import ComposeResult
from textual.containers import Container
from textual.screen import ModalScreen
from textual.widgets import ListItem, ListView


class SelectorScreen(ModalScreen[int]):
    BINDINGS = [('q', 'cancel', 'Cancel')]

    def __init__(self, options: List[ListItem], title: Optional[str] = None):
        super().__init__()
        self.options = options
        self.title = title

    def compose(self) -> ComposeResult:
        with Container(id='selector-dialog'):
            list_view = ListView(*self.options)
            if self.title:
                list_view.border_title = self.title
            yield list_view

    def on_list_view_selected(self, event: ListView.Selected):
        self.dismiss(event.list_view.index)

    def action_cancel(self):
        self.dismiss(None)



import pathlib
from dataclasses import dataclass
from typing import Any, Dict, List, Optional

from rbx import console, utils
from rbx.box import package, presets
from rbx.box.fields import Primitive
from rbx.box.schema import (
    CheckerTest,
    CodeItem,
    ExpectedOutcome,
    Generator,
    GeneratorCall,
    Interactor,
    Package,
    Solution,
    TaskType,
    Testcase,
    TestcaseGroup,
    TestcaseSubgroup,
    ValidatorOutcome,
    ValidatorTest,
)
from rbx.box.testing.testing_preset import TestingPreset
from rbx.box.testing.testing_shared import PathOrStr, TestingShared
from rbx.grading.steps import Evaluation
from rbx.testing_utils import print_directory_tree


@dataclass
class TestcaseArtifacts:
    output: Optional[str] = None
    log: Optional[Evaluation] = None
    interactor_input: Optional[str] = None
    interactor_output: Optional[str] = None
    interactor_pipes: Optional[str] = None


class TestingPackage(TestingShared):
    def __init__(self, root: PathOrStr):
        super().__init__(root)
        self._yml = None

        self.initialize()
        self.preset = self.initialize_preset()

    @property
    def yml_path(self) -> pathlib.Path:
        return self.root / 'problem.rbx.yml'

    def initialize(self):
        if not self.yml_path.exists():
            self.yml_path.parent.mkdir(parents=True, exist_ok=True)
            self.yml_path.touch()
            self.yml_path.write_text(
                utils.model_to_yaml(
                    Package(name='test-problem', timeLimit=1000, memoryLimit=256)
                )
            )

    def initialize_preset(self) -> TestingPreset:
        preset = presets.get_active_preset_or_null(self.root)
        if preset is None:
            preset_path = self.root / '.local.rbx'
            preset_path.mkdir(parents=True, exist_ok=True)
        else:
            preset_path = presets.get_active_preset_path(self.root)
        return TestingPreset(preset_path)

    def print_tree(self):
        print_directory_tree(self.root)

    def print_yml(self):
        console.console.print(self.yml_path.read_text(), highlight=True)

    def print_debug(self):
        self.print_yml()
        self.print_tree()

    @property
    def yml(self) -> Package:
        if self._yml is None:
            self._yml = utils.model_from_yaml(Package, self.yml_path.read_text())
        return self._yml

    def save(self):
        self.yml_path.write_text(utils.model_to_yaml(self.yml))
        # Clear internal cache and package cache to ensure the updated package is loaded fresh
        self._yml = None
        package.clear_package_cache()

    def set_type(self, type: TaskType):
        self.yml.type = type
        self.save()

    def add_solution(
        self,
        path: PathOrStr,
        outcome: ExpectedOutcome,
        language: Optional[str] = None,
    ):
        self.yml.solutions = self.yml.solutions + [
            Solution(path=pathlib.Path(path), language=language, outcome=outcome)
        ]
        self.save()
        return self.add_file(path)

    def add_generator(
        self,
        path: PathOrStr,
        language: Optional[str] = None,
        alias: Optional[str] = None,
        src: Optional[PathOrStr] = None,
    ):
        if alias is not None:
            self.yml.generators = self.yml.generators + [
                Generator(path=pathlib.Path(path), language=language, name=alias)
            ]
        self.save()
        return self.add_file(path, src=src)

    def set_validator(
        self,
        path: PathOrStr,
        language: Optional[str] = None,
        src: Optional[PathOrStr] = None,
    ):
        self.yml.validator = CodeItem(path=pathlib.Path(path), language=language)
        self.save()
        return self.add_file(path, src=src)

    def set_checker(
        self,
        path: PathOrStr,
        language: Optional[str] = None,
        src: Optional[PathOrStr] = None,
    ):
        self.yml.checker = CodeItem(path=pathlib.Path(path), language=language)
        self.save()
        return self.add_file(path, src=src)

    def set_interactor(
        self,
        path: PathOrStr,
        language: Optional[str] = None,
        src: Optional[PathOrStr] = None,
    ):
        self.yml.interactor = Interactor(path=pathlib.Path(path), language=language)
        self.save()
        return self.add_file(path, src=src)

    def set_var(self, name: str, value: Primitive):
        self.yml.vars[name] = value
        self.save()

    def set_vars(self, vars: Dict[str, Primitive]):
        self.yml.vars = vars
        self.save()

    def add_testplan(self, name: str, src: Optional[PathOrStr] = None):
        path = self.add_file(pathlib.Path('testplan') / f'{name}.txt', src)
        return path

    def add_testscript(self, name: str, src: Optional[PathOrStr] = None):
        path = self.add_file(pathlib.Path('testplan') / f'{name}.py', src)
        return path

    def add_testgroup_from_glob(
        self,
        name: str,
        glob: str,
        validator: Optional[PathOrStr] = None,
        extra_validators: Optional[List[PathOrStr]] = None,
    ):
        self.yml.testcases = self.yml.testcases + [
            TestcaseGroup(
                name=name,
                testcaseGlob=glob,
                validator=CodeItem(path=pathlib.Path(validator)) if validator else None,
                extraValidators=[
                    CodeItem(path=pathlib.Path(v)) for v in extra_validators
                ]
                if extra_validators
                else [],
            )
        ]
        self.save()

    def add_testgroup_from_plan(
        self,
        name: str,
        plan: str,
        validator: Optional[PathOrStr] = None,
        extra_validators: Optional[List[PathOrStr]] = None,
    ):
        plan_path = self.add_testplan(name)
        plan_path.write_text(plan)
        self.yml.testcases = self.yml.testcases + [
            TestcaseGroup(
                name=name,
                generatorScript=CodeItem(path=plan_path),
                validator=CodeItem(path=pathlib.Path(validator)) if validator else None,
                extraValidators=[
                    CodeItem(path=pathlib.Path(v)) for v in extra_validators
                ]
                if extra_validators
                else [],
            )
        ]
        self.save()

    def add_testgroup_from_script(
        self,
        name: str,
        script: str,
        validator: Optional[PathOrStr] = None,
        extra_validators: Optional[List[PathOrStr]] = None,
    ):
        script_path = self.add_testscript(name)
        script_path.write_text(script)
        self.yml.testcases = self.yml.testcases + [
            TestcaseGroup(
                name=name,
                generatorScript=CodeItem(path=script_path),
                validator=CodeItem(path=pathlib.Path(validator)) if validator else None,
                extraValidators=[
                    CodeItem(path=pathlib.Path(v)) for v in extra_validators
                ]
                if extra_validators
                else [],
            )
        ]
        self.save()

    def add_testgroup_with_subgroups(
        self,
        name: str,
        subgroups: List[Dict[str, Any]],
        validator: Optional[PathOrStr] = None,
        extra_validators: Optional[List[PathOrStr]] = None,
    ):
        """Add a testgroup with subgroups.

        Args:
            name: Name of the testgroup
            subgroups: List of subgroup definitions, each containing fields like:
                - name: subgroup name
                - generators: list of generator calls
                - testcases: list of testcase objects
                - testcaseGlob: glob pattern
                - generatorScript: generator script path
                - extraValidators: list of extra validators
        """

        subgroup_objects = []
        for subgroup_data in subgroups:
            subgroup_dict = {'name': subgroup_data['name']}

            if 'generators' in subgroup_data:
                subgroup_dict['generators'] = [
                    GeneratorCall(name=gen['name'], args=gen.get('args'))
                    for gen in subgroup_data['generators']
                ]

            if 'testcases' in subgroup_data:
                subgroup_dict['testcases'] = [
                    Testcase(
                        inputPath=pathlib.Path(tc['inputPath']),
                        outputPath=pathlib.Path(tc['outputPath'])
                        if tc.get('outputPath')
                        else None,
                    )
                    for tc in subgroup_data['testcases']
                ]

            if 'testcaseGlob' in subgroup_data:
                subgroup_dict['testcaseGlob'] = subgroup_data['testcaseGlob']

            if 'generatorScript' in subgroup_data:
                subgroup_dict['generatorScript'] = CodeItem(
                    path=pathlib.Path(subgroup_data['generatorScript'])
                )

            if 'extraValidators' in subgroup_data:
                subgroup_dict['extraValidators'] = [
                    CodeItem(path=pathlib.Path(v))
                    for v in subgroup_data['extraValidators']
                ]

            subgroup_objects.append(TestcaseSubgroup(**subgroup_dict))

        self.yml.testcases = self.yml.testcases + [
            TestcaseGroup(
                name=name,
                subgroups=subgroup_objects,
                validator=CodeItem(path=pathlib.Path(validator)) if validator else None,
                extraValidators=[
                    CodeItem(path=pathlib.Path(v)) for v in extra_validators
                ]
                if extra_validators
                else [],
            )
        ]
        self.save()

    def add_testgroup_with_manual_testcases(
        self,
        name: str,
        testcases: List[Dict[str, str]],
        validator: Optional[PathOrStr] = None,
        extra_validators: Optional[List[PathOrStr]] = None,
    ):
        """Add a testgroup with manually defined testcases.

        Args:
            name: Name of the testgroup
            testcases: List of testcase definitions, each containing:
                - inputPath: path to input file
                - outputPath: optional path to output file
        """

        testcase_objects = []
        for tc_data in testcases:
            testcase_objects.append(
                Testcase(
                    inputPath=pathlib.Path(tc_data['inputPath']),
                    outputPath=pathlib.Path(tc_data['outputPath'])
                    if tc_data.get('outputPath')
                    else None,
                )
            )

        self.yml.testcases = self.yml.testcases + [
            TestcaseGroup(
                name=name,
                testcases=testcase_objects,
                validator=CodeItem(path=pathlib.Path(validator)) if validator else None,
                extraValidators=[
                    CodeItem(path=pathlib.Path(v)) for v in extra_validators
                ]
                if extra_validators
                else [],
            )
        ]
        self.save()

    def add_testgroup_with_generators(
        self,
        name: str,
        generators: List[Dict[str, str]],
        validator: Optional[PathOrStr] = None,
        extra_validators: Optional[List[PathOrStr]] = None,
    ):
        """Add a testgroup with generator calls.

        Args:
            name: Name of the testgroup
            generators: List of generator definitions, each containing:
                - name: generator name
                - args: optional generator arguments
        """

        generator_objects = []
        for gen_data in generators:
            generator_objects.append(
                GeneratorCall(name=gen_data['name'], args=gen_data.get('args'))
            )

        self.yml.testcases = self.yml.testcases + [
            TestcaseGroup(
                name=name,
                generators=generator_objects,
                validator=CodeItem(path=pathlib.Path(validator)) if validator else None,
                extraValidators=[
                    CodeItem(path=pathlib.Path(v)) for v in extra_validators
                ]
                if extra_validators
                else [],
            )
        ]
        self.save()

    def get_build_testgroup_path(self, name: str) -> pathlib.Path:
        return self.root / 'build' / 'tests' / name

    def get_testcase_contents(self, path: pathlib.Path) -> TestcaseArtifacts:
        contents = TestcaseArtifacts()
        output_path = path.with_suffix('.out')
        if output_path.exists():
            contents.output = output_path.read_text()
        log_path = path.with_suffix('.log')
        if log_path.exists():
            contents.log = Evaluation.model_validate_json(log_path.read_text())
        interactor_input_path = path.with_suffix('.pin')
        if interactor_input_path.exists():
            contents.interactor_input = interactor_input_path.read_text()
        interactor_output_path = path.with_suffix('.pout')
        if interactor_output_path.exists():
            contents.interactor_output = interactor_output_path.read_text()
        interactor_pipes_path = path.with_suffix('.pio')
        if interactor_pipes_path.exists():
            contents.interactor_pipes = interactor_pipes_path.read_text()
        return contents

    def add_validator_unit_test(
        self,
        glob: str,
        outcome: ValidatorOutcome = ValidatorOutcome.VALID,
        validator: Optional[PathOrStr] = None,
        files: Optional[Dict[str, str]] = None,
    ):
        """Add a unit test for the validator.

        Args:
            glob: Glob pattern for input files
            outcome: Expected validation outcome
            validator: Optional validator to use (if not main validator)
            files: Optional dict of {filename: content} to create test files
        """
        if files:
            for filename, content in files.items():
                self.add_file(filename).write_text(content)

        validator_test = ValidatorTest(
            glob=glob,
            outcome=outcome,
            validator=CodeItem(path=pathlib.Path(validator)) if validator else None,
        )

        # Explicitly set the unitTests field to mark it as dirty
        unit_tests = self.yml.unitTests
        unit_tests.validator = unit_tests.validator + [validator_test]
        self.yml.unitTests = unit_tests
        self.save()

    def add_checker_unit_test(
        self,
        glob: str,
        outcome: ExpectedOutcome = ExpectedOutcome.ACCEPTED,
        files: Optional[Dict[str, str]] = None,
    ):
        """Add a unit test for the checker.

        Args:
            glob: Glob pattern for test files
            outcome: Expected checker outcome
            files: Optional dict of {filename: content} to create test files
        """
        if files:
            for filename, content in files.items():
                self.add_file(filename).write_text(content)

        checker_test = CheckerTest(glob=glob, outcome=outcome)

        # Explicitly set the unitTests field to mark it as dirty
        unit_tests = self.yml.unitTests
        unit_tests.checker = unit_tests.checker + [checker_test]
        self.yml.unitTests = unit_tests
        self.save()



import pathlib
from typing import Any, Dict, Optional

import questionary
import rich
import rich.console
import typer
from ordered_set import OrderedSet
from pydantic import BaseModel, Field

from rbx import console, utils
from rbx.box import environment, limits_info, package, schema
from rbx.box.code import find_language_name
from rbx.box.environment import VerificationLevel
from rbx.box.formatting import href
from rbx.box.schema import ExpectedOutcome
from rbx.box.solutions import (
    RunSolutionResult,
    consume_and_key_evaluation_items,
    get_exact_matching_solutions,
    print_run_report,
    run_solutions,
)


class TimingProfile(BaseModel):
    timeLimit: int
    formula: Optional[str] = None
    timeLimitPerLanguage: Dict[str, int] = Field(default_factory=dict)

    def to_limits(self):
        return schema.LimitsProfile(
            timeLimit=self.timeLimit,
            formula=self.formula,
            modifiers={
                lang: schema.LimitModifiers(time=tl)
                for lang, tl in self.timeLimitPerLanguage.items()
            },
        )


def get_timing_profile(
    profile: str, root: pathlib.Path = pathlib.Path()
) -> Optional[TimingProfile]:
    path = package.get_limits_file(profile, root)
    if not path.exists():
        return None
    return utils.model_from_yaml(TimingProfile, path.read_text())


def step_down(x: Any, step: int) -> int:
    x = int(x)
    return x // step * step


def step_up(x: Any, step: int) -> int:
    x = int(x)
    return (x + step - 1) // step * step


async def estimate_time_limit(
    console: rich.console.Console,
    result: RunSolutionResult,
    formula: Optional[str] = None,
) -> Optional[TimingProfile]:
    structured_evaluations = consume_and_key_evaluation_items(
        result.items, result.skeleton
    )

    timing_per_solution = {}
    timing_per_solution_per_language = {}

    if not result.skeleton.solutions:
        console.print('[error]No solutions to estimate time limit from.[/error]')
        return None

    for solution in result.skeleton.solutions:
        timings = []
        for evals in structured_evaluations[str(solution.path)].values():
            for ev in evals:
                if ev is None:
                    continue
                ev = await ev()
                if ev.log.time is not None:
                    timings.append(int(ev.log.time * 1000))

        if not timings:
            console.print(
                f'[warning]No timings for solution {href(solution.path)}.[/warning]'
            )
            continue

        timing_per_solution[str(solution.path)] = max(timings)
        lang = find_language_name(solution)
        if lang not in timing_per_solution_per_language:
            timing_per_solution_per_language[lang] = {}
        timing_per_solution_per_language[lang][str(solution.path)] = max(timings)

    console.rule('[status]Time report[/status]', style='status')

    fastest_time = min(timing_per_solution.values())
    slowest_time = max(timing_per_solution.values())

    console.print(f'Fastest solution: {fastest_time} ms')
    console.print(f'Slowest solution: {slowest_time} ms')

    def _get_lang_fastest(lang: str) -> int:
        return min(timing_per_solution_per_language[lang].values())

    def _get_lang_slowest(lang: str) -> int:
        return max(timing_per_solution_per_language[lang].values())

    env = environment.get_environment()
    if formula is None:
        formula = env.timing.formula

    def _eval(fastest_time: int, slowest_time: int) -> int:
        return int(
            eval(
                formula,
                {
                    'fastest': fastest_time,
                    'slowest': slowest_time,
                    'step_up': step_up,
                    'step_down': step_down,
                },
            )
        )

    if len(timing_per_solution_per_language) > 1:
        timing_language_list = [
            (_get_lang_fastest(lang), lang) for lang in timing_per_solution_per_language
        ]
        fastest_language_time, fastest_language = min(timing_language_list)
        slowest_language_time, slowest_language = max(timing_language_list)

        console.print(
            f'Fastest language: {fastest_language} ({fastest_language_time} ms)'
        )
        console.print(
            f'Slowest language: {slowest_language} ({slowest_language_time} ms)'
        )

    console.print()
    console.rule('[status]Time estimation[/status]', style='status')

    console.print(f'Using formula: {formula}')

    estimated_tl = _eval(fastest_time, slowest_time)
    console.print(f'[success]Estimated time limit:[/success] {estimated_tl} ms')

    estimated_tl_per_language = {}
    if len(timing_per_solution_per_language) > 1:
        for lang in timing_per_solution_per_language:
            estimated_tl_per_language[lang] = _eval(
                _get_lang_fastest(lang), _get_lang_slowest(lang)
            )

    final_estimated_tls_per_language = {}
    if estimated_tl_per_language:
        for lang, tl in estimated_tl_per_language.items():
            console.print(f'Estimated time limit for {lang}: {tl} ms')

        all_distinct_tls = set(estimated_tl_per_language.values())
        if len(all_distinct_tls) > 1:
            console.print()
            console.print('It seems your problem has solutions for multiple languages!')
            selected_langs = await questionary.checkbox(
                'Please select which languages you want to have a specific time limit for '
                '(or leave all unselected if you want to use a single global time limit)',
                choices=list(estimated_tl_per_language.keys()),
            ).ask_async()
            if selected_langs:
                for lang in selected_langs:
                    final_estimated_tls_per_language[lang] = estimated_tl_per_language[
                        lang
                    ]

    return TimingProfile(
        timeLimit=estimated_tl,
        formula=formula,
        timeLimitPerLanguage=final_estimated_tls_per_language,
    )


async def compute_time_limits(
    check: bool,
    detailed: bool,
    runs: int = 0,
    profile: str = 'local',
    formula: Optional[str] = None,
):
    if package.get_main_solution() is None:
        console.console.print(
            '[warning]No main solution found, so cannot estimate a time limit.[/warning]'
        )
        return None

    verification = VerificationLevel.ALL_SOLUTIONS.value

    with utils.StatusProgress('Running ACCEPTED solutions...') as s:
        tracked_solutions = OrderedSet(
            str(solution.path)
            for solution in get_exact_matching_solutions(ExpectedOutcome.ACCEPTED)
        )
        solution_result = run_solutions(
            progress=s,
            tracked_solutions=tracked_solutions,
            check=check,
            verification=VerificationLevel(verification),
            timelimit_override=-1,  # Unlimited for time limit estimation
            nruns=runs,
        )

    console.console.print()
    console.console.rule(
        '[status]Run report (for time estimation)[/status]', style='status'
    )
    ok = await print_run_report(
        solution_result,
        console.console,
        VerificationLevel(verification),
        detailed=detailed,
        skip_printing_limits=True,
    )

    if not ok:
        console.console.print(
            '[error]Failed to run ACCEPTED solutions, so cannot estimate a reliable time limit.[/error]'
        )
        return None

    estimated_tl = await estimate_time_limit(console.console, solution_result, formula)
    if estimated_tl is None:
        return None

    limits_path = package.get_limits_file(profile)
    console.console.print(
        f'[green]Writing the following timing profile to [item]{href(limits_path)}[/item].[/green]'
    )
    console.console.print(estimated_tl, highlight=True)
    limits_path.write_text(utils.model_to_yaml(estimated_tl.to_limits()))

    return estimated_tl


def inherit_time_limits(profile: str = 'local'):
    limits_path = package.get_limits_file(profile)
    limits = schema.LimitsProfile(inheritFromPackage=True)
    limits_path.write_text(utils.model_to_yaml(limits))

    console.console.print(
        f'[green]Inherit time limits from package for profile [item]{profile}[/item].[/green]'
    )


def set_time_limit(timelimit: int, profile: str = 'local'):
    limits = schema.LimitsProfile(timeLimit=timelimit)
    limits_path = package.get_limits_file(profile)
    limits_path.write_text(utils.model_to_yaml(limits))

    console.console.print(
        f'[green]Set time limit for profile [item]{profile}[/item] to [item]{timelimit} ms[/item].[/green]'
    )


def integrate(profile: str = 'local'):
    limits_profile = limits_info.get_saved_limits_profile(profile)
    if limits_profile is None:
        console.console.print(
            f'[error]No limits profile found for profile [item]{profile}[/item].[/error]'
        )
        raise typer.Exit(1)

    if limits_profile.inheritFromPackage:
        console.console.print(
            f'[warning]Limits profile [item]{profile}[/item] already inherits from package.[/warning]'
        )
        console.console.print('[warning]This operation is a no-op.[/warning]')
        return

    ru, pkg = package.get_ruyaml()

    if limits_profile.timeLimit is not None:
        pkg['timeLimit'] = limits_profile.timeLimit
    if limits_profile.memoryLimit is not None:
        pkg['memoryLimit'] = limits_profile.memoryLimit
    if limits_profile.outputLimit is not None:
        pkg['outputLimit'] = limits_profile.outputLimit

    for lang, limits in limits_profile.modifiers.items():
        if limits.time is not None:
            pkg['modifiers'][lang]['time'] = limits.time
        if limits.memory is not None:
            pkg['modifiers'][lang]['memory'] = limits.memory
        if limits.timeMultiplier is not None:
            pkg['modifiers'][lang]['timeMultiplier'] = limits.timeMultiplier

    dest_yml = package.find_problem_yaml()
    assert dest_yml is not None
    utils.save_ruyaml(dest_yml, ru, pkg)

    console.console.print(
        f'[green]Integrated limits profile [item]{profile}[/item] into package.[/green]'
    )



import functools
import pathlib
from typing import List, Optional, Tuple

import ruyaml
import typer
from pydantic import ValidationError

from rbx import console, utils
from rbx.box import cd
from rbx.box.contest.schema import Contest
from rbx.box.package import find_problem_package_or_die
from rbx.box.sanitizers import issue_stack
from rbx.box.schema import Package

YAML_NAME = 'contest.rbx.yml'


@functools.cache
def find_contest_yaml(root: pathlib.Path = pathlib.Path()) -> Optional[pathlib.Path]:
    root = utils.abspath(root)
    contest_yaml_path = root / YAML_NAME
    while root != pathlib.PosixPath('/') and not contest_yaml_path.is_file():
        root = root.parent
        contest_yaml_path = root / YAML_NAME
    if not contest_yaml_path.is_file():
        return None
    return contest_yaml_path


@functools.cache
def find_contest_package(root: pathlib.Path = pathlib.Path()) -> Optional[Contest]:
    contest_yaml_path = find_contest_yaml(root)
    if not contest_yaml_path:
        return None
    try:
        return utils.model_from_yaml(Contest, contest_yaml_path.read_text())
    except ValidationError as e:
        console.console.print(e)
        console.console.print('[error]Error parsing contest.rbx.yml.[/error]')
        raise typer.Exit(1) from e


def find_contest_package_or_die(root: pathlib.Path = pathlib.Path()) -> Contest:
    package = find_contest_package(root)
    if package is None:
        console.console.print(f'Contest not found in {root.absolute()}', style='error')
        raise typer.Exit(1)
    return package


def find_contest(root: pathlib.Path = pathlib.Path()) -> pathlib.Path:
    found = find_contest_yaml(root)
    if found is None:
        console.console.print(f'Contest not found in {root.absolute()}', style='error')
        raise typer.Exit(1)
    return found.parent


def within_contest(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        with cd.new_package_cd(find_contest()):
            issue_level_token = issue_stack.issue_level_var.set(
                issue_stack.IssueLevel.OVERVIEW
            )
            ret = func(*args, **kwargs)
            issue_stack.print_current_report()
            issue_stack.issue_level_var.reset(issue_level_token)
            return ret

    return wrapper


def save_contest(
    package: Optional[Contest] = None, root: pathlib.Path = pathlib.Path()
) -> None:
    package = package or find_contest_package_or_die(root)
    contest_yaml_path = find_contest_yaml(root)
    if not contest_yaml_path:
        console.console.print(f'Contest not found in {root.absolute()}', style='error')
        raise typer.Exit(1)
    contest_yaml_path.write_text(utils.model_to_yaml(package))


def get_problems(contest: Contest) -> List[Package]:
    problems = []
    for problem in contest.problems:
        problems.append(find_problem_package_or_die(problem.get_path()))
    return problems


def get_ruyaml(root: pathlib.Path = pathlib.Path()) -> Tuple[ruyaml.YAML, ruyaml.Any]:
    contest_yaml_path = find_contest_yaml(root)
    if contest_yaml_path is None:
        console.console.print(f'[error]Contest not found in {root.absolute()}[/error]')
        raise typer.Exit(1)
    res = ruyaml.YAML()
    return res, res.load(contest_yaml_path.read_text())
